"use strict";
/***********************************************************************
 * YoLink Homebridge Platform class
 *
 * Copyright (c) 2022-2023 David Kerr
 *
 * Based on https://github.com/homebridge/homebridge-plugin-template
 *
 * This class is the main constructor for the plugin.
 */
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.YoLinkHomebridgePlatform = void 0;
const settings_1 = require("./settings");
const platformAccessory_1 = require("./platformAccessory");
const yolinkAPI_1 = require("./yolinkAPI");
const semaphore_promise_1 = __importDefault(require("semaphore-promise"));
// eslint-disable-next-line @typescript-eslint/no-var-requires
const packageJSON = require('../package.json');
class YoLinkHomebridgePlatform {
    /*********************************************************************
     * constructor
     *
     */
    constructor(log, config, api) {
        var _a, _b, _c, _d, _e, _f, _g, _h, _j, _k;
        var _l, _m, _o, _p, _q, _r, _s, _t, _u, _v;
        this.log = log;
        this.config = config;
        this.api = api;
        this.Service = this.api.hap.Service;
        this.Characteristic = this.api.hap.Characteristic;
        // this is used to track restored cached accessories
        this.accessories = [];
        this.deviceAccessories = [];
        // track known devices so we can detect new ones.
        this.knownDevices = [];
        this.addDeviceSemaphore = new semaphore_promise_1.default();
        this.initializeCount = 0;
        // We need to serialize requests to YoLink API.  Multiple threads can request state
        // updates for a device at the same time.  This would not be good, so we need a
        // semaphore to make sure we don't send a 2nd request before prior one has completed.
        // using a global semaphore rather than per-device to fix YoLink 000201 errors.
        this.yolinkRequestSemaphore = new semaphore_promise_1.default();
        this.reportError = '\nPlease report all bugs at ' + packageJSON.bugs.url + '\n';
        Error.stackTraceLimit = 100;
        // transforms array of devices into object that can be referenced by deviceId...
        const devices = {};
        if (this.config.devices) {
            this.config.devices.forEach(x => devices[String(x.deviceId).toLowerCase()] = x.config);
        }
        this.config.devices = devices;
        this.config.verboseLog = this.makeBoolean(this.config.verboseLog, false);
        this.config.liteLog = this.makeBoolean(this.config.liteLog, true);
        this.config.allDevices = this.makeBoolean(this.config.allDevices, true);
        (_a = (_l = this.config).excludeTypes) !== null && _a !== void 0 ? _a : (_l.excludeTypes = ['Hub', 'SpeakerHub']);
        (_b = (_m = this.config).includeTypes) !== null && _b !== void 0 ? _b : (_m.includeTypes = []);
        this.config.enableExperimental = this.makeBoolean(this.config.enableExperimental, false);
        this.config.deviceTemperatures = this.makeBoolean(this.config.deviceTemperatures, false);
        (_c = (_o = this.config).powerFailureSensorAs) !== null && _c !== void 0 ? _c : (_o.powerFailureSensorAs = 'Outlet');
        (_d = (_p = this.config).mqttPort) !== null && _d !== void 0 ? _d : (_p.mqttPort = settings_1.YOLINK_MQTT_PORT);
        (_e = (_q = this.config).apiURL) !== null && _e !== void 0 ? _e : (_q.apiURL = settings_1.YOLINK_API_URL);
        (_f = (_r = this.config).tokenURL) !== null && _f !== void 0 ? _f : (_r.tokenURL = settings_1.YOLINK_TOKEN_URL);
        (_g = (_s = this.config).version) !== null && _g !== void 0 ? _g : (_s.version = packageJSON.version);
        (_h = (_t = this.config).garageDoors) !== null && _h !== void 0 ? _h : (_t.garageDoors = []);
        (_j = (_u = this.config).refreshAfter) !== null && _j !== void 0 ? _j : (_u.refreshAfter = settings_1.YOLINK_REFRESH_INTERVAL);
        (_k = (_v = this.config).checkNewDeviceInterval) !== null && _k !== void 0 ? _k : (_v.checkNewDeviceInterval = 0);
        this.log.info(`YoLink plugin for HomeBridge version ${packageJSON.version} (c) 2022-2023 David A. Kerr${this.reportError}`);
        this.verboseLog(`Loaded configuration:\n${JSON.stringify(this.config, null, 2)}`);
        this.yolinkAPI = new yolinkAPI_1.YoLinkAPI(this);
        // When this event is fired it means Homebridge has restored all cached accessories from disk.
        // Dynamic Platform plugins should only register new accessories after this event was fired,
        // in order to ensure they weren't added to homebridge already. This event can also be used
        // to start discovery of new accessories.
        this.api.on('didFinishLaunching', () => {
            this.discoverDevices();
        });
    }
    /*********************************************************************
     * This function is invoked when homebridge restores cached accessories
     * from disk at startup.
     */
    configureAccessory(accessory) {
        this.verboseLog('Loading accessory from cache:' + accessory.displayName);
        // add the restored accessory to the accessories cache so we can track if it has already been registered
        this.accessories.push(accessory);
    }
    /*********************************************************************
     * verboseLog
     */
    verboseLog(msg) {
        if (this.config.verboseLog) {
            this.log.info(`[verbose] ${msg}`);
        }
        else {
            this.log.debug(`[verbose] ${msg}`);
        }
    }
    /*********************************************************************
     * liteLog
     */
    liteLog(msg) {
        if (this.config.liteLog) {
            if (this.config.verboseLog) {
                this.log.info(`${msg} [lite]`);
            }
            else {
                this.log.debug(`${msg} [lite]`);
            }
        }
        else {
            this.log.info(`${msg} [lite]`);
        }
    }
    /*********************************************************************
     * makeBoolean
     * Allow for both 'true' as a boolean and "true" as a string to equal
     * true.  And provide a default for when it is undefined.
     */
    makeBoolean(a, b) {
        return (typeof a === 'undefined') ? b : String(a).toLowerCase() === 'true' || a === true;
    }
    /*********************************************************************
     * discoverDevices
     */
    async discoverDevices() {
        try {
            await this.yolinkAPI.login(this);
            await this.registerDevices();
            await this.registerMqtt();
        }
        catch (e) {
            const msg = (e instanceof Error) ? e.stack : e;
            this.log.error('Fatal error during YoLink plugin initialization:\n' + msg);
        }
    }
    /*********************************************************************
     * registerDevices
     */
    async registerDevices() {
        var _a, _b;
        const deviceList = await this.yolinkAPI.getDeviceList(this);
        this.removeDeletedDevices(deviceList);
        // loop over the discovered devices and register each one
        for (const device of deviceList) {
            this.checkAddDevice(device);
        }
        // Now handle garage doors... two devices bound together. Start by removing
        // any existing garage door accessory that may no longer be configured.
        for (const accessory of this.accessories) {
            const device = accessory.context.device;
            const device2 = accessory.context.device2;
            if (device2) {
                device2.name = (_a = device2.config.name) !== null && _a !== void 0 ? _a : device2.name;
                device2.deviceMsgName = `${device2.modelName} (${device2.deviceId}) ${device2.name}`;
                // Delete any existing garage door accessories that do not exactly match one
                // that we have setup in the config file.
                if (!((_b = this.config.garageDoors) === null || _b === void 0 ? void 0 : _b.some(x => x.controller === device.deviceId && x.sensor === device2.deviceId))) {
                    this.log.warn(`Removing Garage Door accessory from cache: ${accessory.displayName} (${device.deviceId} & ${device2.deviceId})`);
                    this.api.unregisterPlatformAccessories(settings_1.PLUGIN_NAME, settings_1.PLATFORM_NAME, [accessory]);
                }
            }
        }
        // Now add each garage door accessory
        for (const garage of this.config.garageDoors) {
            // check that sensor and controller are in our device list
            const garageDevices = deviceList.filter(x => (x.deviceId === garage.controller || x.deviceId === garage.sensor));
            if (garageDevices.length !== 2) {
                this.log.warn(`Garage Door must have two known devices. Ignoring this door:\n${JSON.stringify(garage, null, 2)}`);
                continue;
            }
            const controller = garageDevices.find(x => x.deviceId === garage.controller);
            const sensor = garageDevices.find(x => x.deviceId === garage.sensor);
            if ((sensor === null || sensor === void 0 ? void 0 : sensor.type) !== 'DoorSensor' || !((controller === null || controller === void 0 ? void 0 : controller.type) === 'Finger' || (controller === null || controller === void 0 ? void 0 : controller.type) === 'GarageDoor')) {
                this.log.warn('Garage Door sensor must be of type \'DoorSensor\' and controller of type \'Finger\' or \'GarageDoor\' ' +
                    `Check config file for deviceID typo. Ignoring this door:\n${JSON.stringify(garage, null, 2)}`);
                continue;
            }
            sensor.timeout = garage.timeout;
            const uuid = this.api.hap.uuid.generate(`${garage.controller}${garage.sensor}`);
            // see if an accessory with the same uuid has already been registered and restored from
            // the cached devices we stored in the `configureAccessory` method above.
            const existingAccessory = this.accessories.find(accessory => accessory.UUID === uuid);
            this.addDevice(controller, uuid, existingAccessory, sensor);
            this.verboseLog(`Door Sensor: ${garage.sensor} linked to Garage Controller: ${garage.controller}`);
        }
        // check for mistyped deviceId's in config file
        for (const [device, info] of Object.entries(this.config.devices)) {
            if (!deviceList.some(x => x.deviceId === device)) {
                this.log.warn(`Device "${device}" does not exist in YoLink device list (${JSON.stringify(info, null, 2)}). ` +
                    'Check config file for deviceID typo.');
            }
        }
        // Add an interval timer to check if device has been added or removed
        if (this.config.checkNewDeviceInterval > 0) {
            this.log.info(`Setting interval timer to check for new devices every ${this.config.checkNewDeviceInterval} seconds`);
            setInterval(async () => {
                this.liteLog(`Check new devices timer fired, next check in ${this.config.checkNewDeviceInterval} seconds`);
                await this.checkNewDevice();
            }, this.config.checkNewDeviceInterval * 1000);
        }
    }
    /*********************************************************************
     * checkAddDevice
     */
    checkAddDevice(device) {
        var _a, _b, _c;
        this.verboseLog(JSON.stringify(device, null, 2));
        // Track all devices by ID
        this.knownDevices.push(device.deviceId);
        // Get the config file settings for this device (if set)
        device.config = (_a = this.config.devices[device.deviceId]) !== null && _a !== void 0 ? _a : {};
        device.name = (_b = device.config.name) !== null && _b !== void 0 ? _b : device.name;
        device.deviceMsgName = `${device.modelName} (${device.deviceId}) ${device.name}`;
        // If device is assigned to a garage door then hide it as we will
        // handle those as special case.
        const garage = (_c = this.config.garageDoors) === null || _c === void 0 ? void 0 : _c.some(x => (x.sensor === device.deviceId || x.controller === device.deviceId));
        // Skip over devices that are marked to as hide true, or the device type is listed in the excludeTypes or
        // includeTypes array based on whether allDevices is set to true or false...
        const skip = this.makeBoolean(device.config.hide, !((this.config.allDevices && this.config.excludeTypes.some(x => (x === device.type))) !==
            (this.config.allDevices || this.config.includeTypes.some(x => (x === device.type)))));
        // Generate a unique id for the accessory
        const uuid = this.api.hap.uuid.generate(device.deviceId);
        // See if an accessory with the same uuid has already been registered and restored from
        // the cached devices we stored in the `configureAccessory` method above.
        const existingAccessory = this.accessories.find(accessory => accessory.UUID === uuid);
        // Now add the device...
        if (skip || garage) {
            if (existingAccessory) {
                this.log.warn(`[${device.deviceMsgName}] Remove accessory from cache as ${(garage)
                    ? 'device assigned to garage door' : 'config \'hide=true\''}`);
                this.api.unregisterPlatformAccessories(settings_1.PLUGIN_NAME, settings_1.PLATFORM_NAME, [existingAccessory]);
            }
            else {
                if (garage) {
                    this.log.info(`[${device.deviceMsgName}] Assigned to a garage door`);
                    this.initializeCount++;
                    this.log.debug(`[${device.deviceMsgName}] Increment initialize count: ${this.initializeCount}`);
                }
                else {
                    this.log.info(`[${device.deviceMsgName}] Not registering device as config 'hide=true'`);
                }
            }
        }
        else {
            this.initializeCount++;
            this.log.debug(`[${device.deviceMsgName}] Increment initialize count: ${this.initializeCount}`);
            this.addDevice(device, uuid, existingAccessory);
        }
    }
    /*********************************************************************
     * addDevice
     */
    addDevice(device, uuid, existingAccessory, device2 = undefined) {
        if (existingAccessory) {
            // update existing accessory
            this.verboseLog(`[${device.deviceMsgName}] Restoring accessory from cache`);
            existingAccessory.context.device = device;
            existingAccessory.context.device2 = device2;
            this.api.updatePlatformAccessories([existingAccessory]);
            this.deviceAccessories.push(new platformAccessory_1.YoLinkPlatformAccessory(this, existingAccessory));
            return (existingAccessory);
        }
        else {
            // create a new accessory
            this.log.info(`[${device.deviceMsgName}] Adding new accessory`);
            const accessory = new this.api.platformAccessory(device.name, uuid);
            accessory.context.device = device;
            accessory.context.device2 = device2;
            this.api.registerPlatformAccessories(settings_1.PLUGIN_NAME, settings_1.PLATFORM_NAME, [accessory]);
            this.deviceAccessories.push(new platformAccessory_1.YoLinkPlatformAccessory(this, accessory));
            return (accessory);
        }
    }
    /*********************************************************************
     * deleteRemovedDevices
     */
    removeDeletedDevices(deviceList) {
        // Remove accessories from cache if they are no longer in list of
        // devices retrieved from YoLink.
        for (let i = this.accessories.length - 1; i >= 0; i--) {
            const accessory = this.accessories[i];
            const device = accessory.context.device;
            if (!deviceList.some(x => x.deviceId === device.deviceId)) {
                this.log.warn(`Removing accessory from cache: ${accessory.displayName} (${device.deviceId}), device does not exist`);
                const indexKnownDevices = this.knownDevices.indexOf(device.deviceId);
                if (indexKnownDevices >= 0) {
                    this.knownDevices.splice(indexKnownDevices, 1);
                }
                this.accessories.splice(i, 1);
                this.api.unregisterPlatformAccessories(settings_1.PLUGIN_NAME, settings_1.PLATFORM_NAME, [accessory]);
            }
        }
    }
    /*********************************************************************
     * checkNewDevices
     */
    async checkNewDevice(deviceId = undefined) {
        // serialize adding new devices
        const releaseSemaphore = await this.addDeviceSemaphore.acquire();
        try {
            if (deviceId && this.knownDevices.includes(deviceId)) {
                // If deviceId already a known device then return
                return false;
            }
            const deviceList = await this.yolinkAPI.getDeviceList(this);
            this.verboseLog(`Known Device List: \n${this.knownDevices}`);
            this.removeDeletedDevices(deviceList);
            // loop over the discovered devices and register each one if it has not already been registered
            for (const device of deviceList) {
                if (!this.knownDevices.includes(device.deviceId)) {
                    if (deviceId === undefined || deviceId === device.deviceId) {
                        this.checkAddDevice(device);
                    }
                }
            }
            return (true);
        }
        catch (e) {
            const msg = (e instanceof Error) ? e.stack : e;
            this.log.error(`Fatal error checkNewDevices:\n${msg}`);
            return (false);
        }
        finally {
            releaseSemaphore();
        }
    }
    /*********************************************************************
     * registerMqtt
     */
    async registerMqtt() {
        // Now connect to YoLink MQTT server and subscribe to messages
        this.yolinkAPI.mqtt(this, async (message) => {
            // This function is called for every message received over MQTT
            const data = JSON.parse(message);
            // Find the device in the deviceAccessories list
            const deviceAccessory = this.deviceAccessories.find(x => x.deviceId === data.deviceId || x.deviceId2 === data.deviceId);
            // pass the message on to the appropriate device accessory if it exists.
            if (deviceAccessory) {
                deviceAccessory.mqttMessage(data);
            }
            else {
                this.verboseLog(`Known Device List: \n${this.knownDevices}`);
                if (await this.checkNewDevice(data.deviceId)) {
                    // New device added
                    this.log.warn(`New device detected: MQTT received ${data.event} message for (${data.deviceId})`);
                }
                else {
                    // If a device is hidden (not loaded into homebridge) then we may receive
                    // messages for it... which is perfectly okay, but worth logging (only if not lite-logging)
                    this.liteLog(`MQTT received ${data.event} message for hidden device (${data.deviceId})`);
                }
            }
        });
    }
}
exports.YoLinkHomebridgePlatform = YoLinkHomebridgePlatform;
//# sourceMappingURL=platform.js.map