"use strict";
/***********************************************************************
 * YoLink temperature / humidity sensor device support
 *
 * Copyright (c) 2022-2023 David Kerr
 *
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.mqttThermoHydroDevice = exports.initThermoHydroDevice = void 0;
/***********************************************************************
 * initThermoHydroDevice
 * Initialize the temperature and humidity device services.
 */
async function initThermoHydroDevice() {
    var _a, _b;
    const platform = this.platform;
    const accessory = this.accessory;
    const device = accessory.context.device;
    if (String(device.config.hide).toLowerCase() === 'thermo') {
        platform.log.info(`[${device.deviceMsgName}] Hide Thermometer service because config.[${device.deviceId}].hide is set to "thermo"`);
        accessory.removeService(accessory.getService(platform.Service.TemperatureSensor));
    }
    else {
        // Not trying to hide the thermometer service.
        this.thermoService = accessory.getService(platform.Service.TemperatureSensor)
            || accessory.addService(platform.Service.TemperatureSensor);
        this.thermoService.setCharacteristic(platform.Characteristic.Name, device.name);
    }
    if (String(device.config.hide).toLowerCase() === 'hydro') {
        platform.log.info(`[${device.deviceMsgName}] Hide Hydrometer service because config.[${device.deviceId}].hide is set to "hydro"`);
        accessory.removeService(accessory.getService(platform.Service.HumiditySensor));
    }
    else {
        // Not trying to hide the hydrometer service.
        this.hydroService = accessory.getService(platform.Service.HumiditySensor)
            || accessory.addService(platform.Service.HumiditySensor);
        this.hydroService.setCharacteristic(platform.Characteristic.Name, device.name);
    }
    // Call get handler to initialize data fields to current state and set
    // timer to regularly update the data.
    await this.refreshDataTimer(handleGetBlocking.bind(this, 'both'));
    // Once we have initial data, setup all the Homebridge handlers
    (_a = this.thermoService) === null || _a === void 0 ? void 0 : _a.getCharacteristic(platform.Characteristic.CurrentTemperature).onGet(handleGet.bind(this, 'thermo'));
    (_b = this.hydroService) === null || _b === void 0 ? void 0 : _b.getCharacteristic(platform.Characteristic.CurrentRelativeHumidity).onGet(handleGet.bind(this, 'hydro'));
}
exports.initThermoHydroDevice = initThermoHydroDevice;
/***********************************************************************
 * handleGet
 *
 * This is an example of JSON object returned.
 * The only alarm state we can act on is the low battery, HomeKit does not
 * have alarm/alert state for high/low temperature or humidity, that would
 * be handled internally within HomeKit based on data returned.
 *
 *  {
 *    "online": true,
 *    "state": {
 *      "alarm": {
 *        "lowBattery": false,
 *        "lowTemp": false,
 *        "highTemp": false,
 *        "lowHumidity": false,
 *        "highHumidity": false,
 *        "period": false,
 *        "code": 0
 *      },
 *      "battery": 4,
 *      "batteryType": "Li",
 *      "humidity": 47.2,
 *      "humidityCorrection": 0,
 *      "humidityLimit": {
 *        "max": 50,
 *        "min": 0
 *      },
 *      "interval": 10,
 *      "mode": "f",
 *      "state": "normal",
 *      "tempCorrection": 0,
 *      "tempLimit": {
 *        "max": 35,
 *        "min": 18
 *      },
 *      "temperature": 25.4,
 *      "version": "0508"
 *    },
 *    "deviceId": "abcdef1234567890",
 *    "reportAt": "2022-07-22T15:32:15.899Z"
 *  }
 */
async function handleGet(sensor = 'thermo') {
    var _a, _b, _c, _d, _e, _f;
    // wrapping the semaphone blocking function so that we return to Homebridge immediately
    // even if semaphore not available.
    const platform = this.platform;
    const device = this.accessory.context.device;
    handleGetBlocking.bind(this, sensor)()
        .then((v) => {
        if (sensor === 'hydro') {
            this.hydroService.updateCharacteristic(platform.Characteristic.CurrentRelativeHumidity, v);
        }
        else {
            this.thermoService.updateCharacteristic(platform.Characteristic.CurrentTemperature, v);
        }
    });
    // Return current state of the device pending completion of the blocking function
    return ((sensor === 'hydro')
        ? ((_c = (_b = (_a = device.data) === null || _a === void 0 ? void 0 : _a.state) === null || _b === void 0 ? void 0 : _b.humidity) !== null && _c !== void 0 ? _c : 0)
        : ((_f = (_e = (_d = device.data) === null || _d === void 0 ? void 0 : _d.state) === null || _e === void 0 ? void 0 : _e.temperature) !== null && _f !== void 0 ? _f : -270));
}
async function handleGetBlocking(sensor = 'thermo') {
    var _a, _b, _c, _d, _e, _f, _g, _h;
    const platform = this.platform;
    const device = this.accessory.context.device;
    // serialize access to device data.
    const releaseSemaphore = await device.semaphore.acquire();
    // 'hydro' or 'thermo' use -270 as the minimum accepted value for temperature default
    let rc = (sensor === 'hydro') ? 0 : -270;
    try {
        if (await this.checkDeviceState(platform, device) && device.data.online) {
            this.logDeviceState(device, `Temperature ${device.data.state.temperature}\u00B0C ` +
                `(${(device.data.state.temperature * 9 / 5 + 32).toFixed(1)}\u00B0F), Humidity ${device.data.state.humidity}, ` +
                `Battery: ${device.data.state.battery} (Requested: ${sensor})`);
            (_a = this.thermoService) === null || _a === void 0 ? void 0 : _a.updateCharacteristic(platform.Characteristic.StatusActive, true);
            (_b = this.thermoService) === null || _b === void 0 ? void 0 : _b.updateCharacteristic(platform.Characteristic.StatusFault, false);
            (_c = this.hydroService) === null || _c === void 0 ? void 0 : _c.updateCharacteristic(platform.Characteristic.StatusActive, true);
            (_d = this.hydroService) === null || _d === void 0 ? void 0 : _d.updateCharacteristic(platform.Characteristic.StatusFault, false);
            if (device.data.state.alarm.lowBattery) {
                platform.log.warn(`Device ${device.deviceMsgName} reports low battery`);
            }
            rc = (sensor === 'hydro') ? device.data.state.humidity : device.data.state.temperature;
        }
        else {
            platform.log.error(`Device offline or other error for ${device.deviceMsgName}`);
            (_e = this.thermoService) === null || _e === void 0 ? void 0 : _e.updateCharacteristic(platform.Characteristic.StatusActive, false);
            (_f = this.thermoService) === null || _f === void 0 ? void 0 : _f.updateCharacteristic(platform.Characteristic.StatusFault, true);
            (_g = this.hydroService) === null || _g === void 0 ? void 0 : _g.updateCharacteristic(platform.Characteristic.StatusActive, false);
            (_h = this.hydroService) === null || _h === void 0 ? void 0 : _h.updateCharacteristic(platform.Characteristic.StatusFault, true);
        }
    }
    catch (e) {
        const msg = (e instanceof Error) ? e.stack : e;
        platform.log.error('Error in ThermoHydroDevice handleGet' + platform.reportError + msg);
    }
    finally {
        releaseSemaphore();
    }
    return (rc);
}
/***********************************************************************
 * mqttThermoHydroDevice
 * Handle message received from MQTT server.
 *
 * This is an example of JSON object returned.
 * The only alarm state we can act on is the low battery, HomeKit does not
 * have alarm/alert state for high/low temperature or humidity, that would
 * be handled internally within HomeKit based on data returned.
 *  {
 *    "event": "THSensor.Report",
 *    "time": 1658507535922,
 *    "msgid": "1658507535922",
 *    "data": {
 *      "state": "normal",
 *      "alarm": {
 *        "lowBattery": false,
 *        "lowTemp": false,
 *        "highTemp": false,
 *        "lowHumidity": false,
 *        "highHumidity": false,
 *        "period": false,
 *        "code": 0
 *      },
 *      "battery": 4,
 *      "mode": "f",
 *      "interval": 10,
 *      "temperature": 25.5,
 *      "humidity": 48.1,
 *      "tempLimit": {
 *        "max": 35,
 *        "min": 18
 *      },
 *      "humidityLimit": {
 *        "max": 50,
 *        "min": 0
 *      },
 *      "tempCorrection": 0,
 *      "humidityCorrection": 0,
 *      "version": "0508",
 *      "batteryType": "Li",
 *      "loraInfo": {
 *        "signal": -64,
 *        "gatewayId": "abcdef1234567890",
 *        "gateways": 1
 *      }
 *    },
 *    "deviceId": "abcdef1234567890"
 *  }
 *
 * The newer X3 sensor also sends a DataRecord which we will quietly ignore
 * {
 *   "event": "THSensor.DataRecord",
 *   "time": 1665651601414,
 *   "msgid": "1665651601414",
 *   "data": {
 *     "records": [{
 *       "temperature": 29.2,
 *       "humidity": 63.9,
 *       "time": "2022-10-13T08:50:01.000Z"
 *     }]
 *   },
 *   "deviceId": "abcdef1234567890"
 * }
 */
async function mqttThermoHydroDevice(message) {
    var _a, _b;
    const platform = this.platform;
    const device = this.accessory.context.device;
    // serialize access to device data.
    const releaseSemaphore = await device.semaphore.acquire();
    try {
        device.updateTime = Math.floor(new Date().getTime() / 1000) + device.config.refreshAfter;
        const mqttMessage = `MQTT: ${message.event} for device ${device.deviceMsgName}`;
        const event = message.event.split('.');
        const batteryMsg = (device.hasBattery && message.data.battery) ? `, Battery: ${message.data.battery}` : '';
        const alertMsg = (message.data.alertType) ? `, Alert: ${message.data.alertType}` : '';
        switch (event[1]) {
            case 'Alert':
            // I can see no way in HomeKit documentation for a thermo/hydro sensor
            // to generate an alert.  I think bounds testing / alerting all has to be
            // handled within HomeKit.
            // falls through
            case 'Report':
                if (!device.data) {
                    // in rare conditions (error conditions returned from YoLink) data object will be undefined or null.
                    platform.log.warn(`Device ${device.deviceMsgName} has no data field, is device offline?`);
                    break;
                }
                // if we received a message then device must be online
                device.data.online = true;
                // Merge received data into existing data object
                Object.assign(device.data.state, message.data);
                if (!message.data.reportAt) {
                    // mqtt data does not include a report time, so merging the objects leaves current
                    // unchanged, update the time string.
                    device.data.reportAt = device.reportAtTime.toISOString();
                }
                this.logDeviceState(device, `Temperature ${message.data.temperature}\u00B0C ` +
                    `(${(message.data.temperature * 9 / 5 + 32).toFixed(1)}\u00B0F), ` +
                    `Humidity ${message.data.humidity}${alertMsg}${batteryMsg} (MQTT: ${message.event})`);
                (_a = this.thermoService) === null || _a === void 0 ? void 0 : _a.updateCharacteristic(platform.Characteristic.CurrentTemperature, message.data.temperature);
                (_b = this.hydroService) === null || _b === void 0 ? void 0 : _b.updateCharacteristic(platform.Characteristic.CurrentRelativeHumidity, message.data.humidity);
                if (message.data.alarm.lowBattery) {
                    platform.log.warn(`Device ${device.deviceMsgName} reports low battery`);
                }
                break;
            case 'DataRecord':
            // falls through
            case 'setAlarm':
                // No equivalent for this in HomeKit
                platform.liteLog(mqttMessage + ' ' + JSON.stringify(message, null, 2));
                break;
            default:
                platform.log.warn(mqttMessage + ' not supported.' + platform.reportError + JSON.stringify(message, null, 2));
        }
    }
    catch (e) {
        const msg = (e instanceof Error) ? e.stack : e;
        platform.log.error('Error in mqttThermoHydroDevice' + platform.reportError + msg);
    }
    finally {
        releaseSemaphore();
    }
}
exports.mqttThermoHydroDevice = mqttThermoHydroDevice;
//# sourceMappingURL=thermoHydroDevice.js.map